'''

.. Delete files when the disk is full
.. 
.. root_dir: /data - contains audio, spl, logs
.. threshold: 0.95
.. batch_size: 1
.. decision:
..  - {dir: logs, }
.. 
.. 
.. 
.. DiskMonitor('/data', [
..     ('logs', batch(n=None)),
..     ('spl', delete_batch(even_distributed, n=5)),
.. ])

'''
import os
import random
import glob
import reip


class DiskMonitor(reip.Block):
    '''Monitors the disk usage in a certain directory and will selectively delete files from that directory.
    
    .. code-block:: python

        # define the file deleter

        @B.diskmonitor.Diskmonitor.deleter
        def Diskmon(block, offset=1, skip=3, chunksize=5):
            # select some subset of the files
            files = block.get_files('audio')[offset::skip]
            # randomly delete files in chunks of 5 until 
            # under the desired storage limit
            block.delete_while_full(files, chunksize)

        # instantiate the block
        with reip.Graph():
            Diskmon('/data')

        
    '''
    def __init__(self, root='/', deleter=None, threshold=0.95, padding=0.1, interval=60, **kw):
        self._deleter = deleter if callable(deleter) else self._default_deleter
        self.root = root
        self.threshold = threshold
        self.padding = padding
        self._files = []
        super().__init__(n_inputs=0, max_rate=1./interval, extra_kw=True, **kw)

    def process(self, *files, meta):
        # initial usage check
        start_usage = self.get_usage()
        if start_usage < self.threshold:
            return

        # ok so we've exceeded, delete some files
        self._files.clear()
        self.log.warning('Disk usage exceeded ({:.1%} > {:.1%}).'.format(start_usage, self.threshold))
        self._deleter(self, **self.extra_kw)

        # send deleted files with change in usage
        usage = self.get_usage()
        self.log.info('Removed {} files. Usage at {:.1%}.'.format(len(self._files), usage))
        return [self._files], {
            'start_usage': start_usage,
            'end_usage': usage,
            'usage_delta': start_usage - usage}

    def get_usage(self):
        return reip.util.status.storage(self.root, literal_keys=True)[self.root] / 100.

    def get_files(self, *fs):
        return [
            f #for root in self._root_dirs
            for f in glob.glob(os.path.join(self.root, *fs), recursive=True)
        ]

    def delete_while_full(self, fs, chunksize=1, method='random'):
        chunksize = chunksize or len(fs)
        if method == 'random':
            random.shuffle(fs)
        elif method == 'newest':
            fs = fs[::-1]
        elif method == 'oldest':
            pass
        i = -chunksize
        for usage, i in zip(self.while_full(), range(0, len(fs), chunksize)):
            self.delete(fs[i:i+chunksize])
        return i + chunksize < len(fs)

    def while_full(self, threshold=None):
        threshold = threshold if threshold is not None else (self.threshold - self.padding)
        while True:
            usage = self.get_usage()
            if usage < threshold:
                break
            yield usage

    def delete(self, fs):
        fs = reip.util.as_list(fs)
        for f in fs:
            os.remove(f)
        self._files.extend(fs)

    @classmethod
    def deleter(cls, *a, **kw):
        return (
            # called deco with no arguments, just the decoed function
            reip.util.partial(cls, *a[1:], deleter=a[0], **kw)
            if a and callable(a[0]) else
            # called with misc arguments, return a function that will create
            # a partial for catching the function
            reip.util.create_partial(cls, *a, **kw)
        )

    @staticmethod
    def _default_deleter(block):
        return block.delete_while_full(block.get_files())




# @DiskMonitor.deleter
# def SonycDiskMonitor(block, chunksize=5, skip=2, offset=1):
#     block.delete(block.get_files('logs'))  # delete all logs first
#     return (
#         block.delete_while_full(block.get_files('audio')[offset::skip], chunksize) or
#         block.delete_while_full(block.get_files('ml')[offset::skip], chunksize) or
#         block.delete_while_full(block.get_files('spl')[offset::skip], chunksize))



import os
import time
import numpy as np
import pyarrow as pa
import pyarrow.plasma as plasma
from .base import BaseStore
from .pointers import SharedPointer
from . import queue as q_


__all__ = ['PlasmaStore', 'ArrowQueue']


def get_plasma_path():
    return os.path.join(os.getenv('TMPDIR') or '/tmp', 'plasma')

# def start_plasma(plasma_socket=None, plasma_store_memory=1e9):
#     import subprocess
#     command = [
#         'plasma_store',
#         '-s', plasma_socket or get_plasma_path(),
#         '-m', str(int(plasma_store_memory))]
#     print('Starting plasma server...', command)
#     proc = subprocess.Popen(command, stdout=None, stderr=None, shell=False)
#     time.sleep(0.5)
#     return proc
#
#
# def connect_to_plasma(plasma_socket=None, **kw):
#     plasma_socket = plasma_socket or get_plasma_path()
#     client = None #plasma.connect(plasma_dir)
#     if client is None:
#         # if getattr(connect_to_plasma, 'proc', None) is None:
#         #     connect_to_plasma.proc = start_plasma(plasma_socket, **kw)
#
#         client = plasma.connect(plasma_socket)
#     return client



def random_object_id():
    return plasma.ObjectID(np.random.bytes(20))

def random_unique_object_id(existing):
    while True:
        id = random_object_id()
        if id not in existing:
            existing.add(id)
            return id

def n_random_unique_object_ids(client, size=10):
    existing_ids = set(client.list())
    return [
        random_unique_object_id(existing_ids)
        for i in range(size)]


class PlasmaStore(BaseStore):
    '''Store that puts data in plasma store.
    
    Requires running a plasma store process.

    e.g.

    .. code-block:: bash

        plasma_store -s $TMPDIR/plasma -m 2000000000
    '''
    Pointer = SharedPointer
    def __init__(self, size, plasma_socket=None):
        self._plasma_socket_name = plasma_socket or get_plasma_path()
        self.client = plasma.connect(self._plasma_socket_name)

        # First write can be very slow on platform like Jetson Nano
        print("Connected to %s. Warming up..." % self._plasma_socket_name)
        t0 = time.time()
        ret = self.client.get(self.client.put("warm-up"))
        assert (ret == "warm-up")
        print("Warmed up in %.4f sec\n" % (time.time()- t0))

        self.ids = n_random_unique_object_ids(self.client, size)
        self.size = size

    _refreshed = False
    def refresh(self):  # TODO: HOW TO CALL THIS WHEN FORKING ???
        self.client = plasma.connect(self._plasma_socket_name)

    def __len__(self):
        return len(self.client.list())

    def put(self, data, meta=None, id=None):
        return save_both(self.client, data, meta or {}, id=self.ids[id])

    def get(self, id):
        if not self._refreshed:
            self.refresh()
            self._refreshed = True
        return load_both(self.client, self.ids[id])

    def delete(self, ids):
        self.client.delete([self.ids[id] for id in ids])


TYPE = '__PLASMA_SERIALIZE_DTYPE__'

def save_both(client, data, meta=None, id=None):
    meta = dict(meta) if meta else {}  # make a copy, don't modify the original dict

    if data is None:
        meta[TYPE] = "void"
        object_size = 0
    elif isinstance(data, str):
        meta[TYPE] = "string"
        data = data.encode("utf-8")
        object_size = len(data)
    elif isinstance(data, np.ndarray):
        meta[TYPE] = "array"
        tensor = pa.Tensor.from_numpy(data)
        object_size = pa.ipc.get_tensor_size(tensor)
    else:
        data = pa.serialize(data).to_buffer()
        object_size = len(data)

    meta = pa.serialize(meta).to_buffer().to_pybytes()

    object_id = id or random_object_id()

    buf = client.create(object_id, object_size, metadata=meta)
    stream = pa.FixedSizeBufferWriter(buf)
    if isinstance(data, (bytes, pa.lib.Buffer)):
        stream.write(data)
    elif data is not None:
        stream.set_memcopy_threads(4)
        pa.ipc.write_tensor(tensor, stream)
    client.seal(object_id)
    return object_id


def load_both(client, id):
    meta, data = client.get_buffers([id], timeout_ms=1, with_meta=True)[0]
    meta = pa.deserialize(meta)
    dtype = meta.pop(TYPE, None)
    if dtype == "void":
        data = None
    elif dtype == "string":
        data = data.to_pybytes().decode("utf-8")
    elif dtype == "array":
        reader = pa.BufferReader(data)
        tensor = pa.ipc.read_tensor(reader)
        data = tensor.to_numpy()
    else:
        data = pa.deserialize(data)
    return data, meta


class pyarrow_serializer:
    def loads(self, obj):
        return pa.deserialize(obj)

    def dumps(self, obj):
        return pa.serialize(obj).to_buffer()
q_.register_serializer(pyarrow_serializer(), 'arrow')

class ArrowQueue(q_.Queue):
    serializer = 'arrow'

#
#
# def test():
#     print(get_plasma_path())
#     client = plasma.connect(get_plasma_path())
#     print(client.store_capacity())
#     print(client.list())
#     client.delete(client.list())
#     print(client.list())
#
#     print("Generating...")
#     data = np.ones(1 * 10 ** 4, dtype=np.uint8)
#     # data = "Hello"
#     # data = None
#     print("Done")
#
#     # id = save_data(client, data)
#     # data2 = load_data(client, id)
#     # print(data2)
#     # client.delete([id])
#
#     meta = {"Foo": 10}
#
#     # id2 = save_meta(client, meta)
#     # meta2 = load_meta(client, id2)
#     # print(meta2)
#
#     id3 = save_both(client, data, meta)
#     data3, meta3 = load_both(client, id3)
#     print(data3, meta3)
#
#
# def test2():
#     store = Store()
#     for i in range(5):
#         print('-'*10, i)
#         data = np.ones(1 * 10 ** 4, dtype=np.uint8)
#         meta = {"Foo": 10}
#
#         id = store.put(data, meta)
#         data2, meta2 = store.get(id)
#         print(id, data2, meta2)
#
#
# if __name__ == '__main__':
#     # test()
#     test2()
#     exit(0)
#
#     client = plasma.connect(get_plasma_path())
#     print(client.store_capacity())
#     print(client.list())
#     client.delete(client.list())
#     print(client.list())
#     client.evict(10000000000)
#
#     object_id = plasma.ObjectID(20 * b"a")
#     object_size = 1000
#     print(object_id)
#     if client.contains(object_id):
#         client.delete([object_id])
#
#     buffer = memoryview(client.create(object_id, object_size))
#
#     for i in range(1000):
#         buffer[i] = i % 128
#     client.seal(object_id)
#
#     print("Generating...")
#     data = np.ones(10**9)
#     print("Done")
#     tensor = pa.Tensor.from_numpy(data)
#
#     random_id = plasma.ObjectID(np.random.bytes(20))
#     sz = pa.ipc.get_tensor_size(tensor)
#     print("sz", sz)
#     buf = client.create(random_id, sz)
#
#     stream = pa.FixedSizeBufferWriter(buf)
#     stream.set_memcopy_threads(6)
#     a = time.time()
#     print("Writing...")
#     pa.ipc.write_tensor(tensor, stream)
#     print("Writing took ", time.time() - a)
#     client.seal(random_id)
#
#     #################################################################
#     client2 = plasma.connect(get_plasma_path())
#
#     t0 = time.time()
#     print("Reading...")
#     [buf2] = client.get_buffers([random_id])
#     reader = pa.BufferReader(buf2)
#     tensor2 = pa.ipc.read_tensor(reader)
#     array = tensor2.to_numpy()
#     # array[10] = 0
#     # array2 = np.copy(array)
#     # array2 = copy.deepcopy(array)
#     # array2[10] = 0
#     print("Reading took ", time.time() - t0)
#
#     object_id2 = plasma.ObjectID(20 * b"a")
#     [buffer2] = client2.get_buffers([object_id2], timeout_ms=1)
#     view2 = memoryview(buffer2)
#     for i in range(1000):
#         if buffer[i] != view2[i]:
#             print("Mismatch", i)
#
#     meta = {"shape": (10, 20), "timestamps": [0, 1, 2], "other": 1.2}
#
#     auto_id = client.put(meta)
#     print(auto_id)
#     got = client.get(auto_id, timeout_ms=1)
#     print(got)



import multiprocessing as mp


class Counter:
    def __init__(self, initial=0):
        self.value = initial

    def as_basic(self):
        return Counter(self.value)

    def as_shared(self):
        return SharedCounter(self.value)

class SharedCounter:
    def __init__(self, initial=0):
        self._value = mp.Value('i', initial, lock=False)
        super().__init__(initial)

    @property
    def value(self):
        return self._value.value

    @value.setter
    def value(self, new_value):
        self._value.value = new_value


class Pointer:
    def __init__(self, size, counter=0):
        self.counter = counter
        self.size = size

    def __str__(self):
        return (
            f'<{self.__class__.__name__} i={self.counter} '
            f'pos={self.pos}/{self.size - 1} loop={self.loop}>')

    @property
    def pos(self):
        return self.counter % self.size

    @property
    def loop(self):
        return self.counter // self.size

    def as_basic(self):
        return Pointer(self.size, self.counter)

    def as_shared(self):
        return SharedPointer(self.size, self.counter)


class SharedPointer(Pointer):
    def __init__(self, size, counter=0):
        self._counter = mp.Value('i', counter, lock=False)
        super().__init__(size, counter)

    @property
    def counter(self):
        return self._counter.value

    @counter.setter
    def counter(self, new_value):
        self._counter.value = new_value



'''


reip.blocks.Lambda(
    reip.util.mergedict(lambda: (
        reip.status.base,
        reip.status.cpu,
        reip.status.memory,
        reip.status.network,
        reip.status.wifi,
    ))
)


'''
import os
import re
import functools
import fnmatch
import socket
from datetime import datetime
import psutil
import ifcfg
import ixconfig
import netswitch
import netswitch.cell
import reip
import logging

log = logging.getLogger(__name__)


STATS_FUNCTIONS = {}
def register_stats(func):
    @functools.wraps(func)
    def stats(*a, **kw):
        try:
            return func(*a, **kw) or {}
        except Exception as e:
            log.exception(e)
            log.error('Error getting {} status: ({}) {}'.format(
                func.__name__, type(e).__name__, e))
        return {}
    STATS_FUNCTIONS[func.__name__] = stats
    return stats


# strip suffix from string if present
stripsfx = lambda x, sfx: x[:-len(sfx)] if x.endswith(sfx) else x
# find regex pattern in shell output
shfind = lambda pat, cmd: re.findall(pat, reip.util.shell.run(cmd)[0])
#
as_kw = lambda kw, key: kw if isinstance(kw, dict) else {key: kw}

# class Status(reip.Block):
#     def __init__(self, **kw):
#         super().__init__(**kw)
#
#     def process(self, meta):
#         data = {}
#         for key, func in STATS_FUNCTIONS.items():
#             try:
#                 data.update(func())
#             except Exception:
#                 self.log.exception()
#         return [data], {}
#


@register_stats
def cpu(meta=None):
    cpu_cur_freq = shfind(r'\b[\d]+', 'vcgencmd measure_clock arm')[-1].split()[0]
    cpu_temp = shfind(r'\b[\d?(.\d)]+\b', 'vcgencmd measure_temp')[0].split()[0]
    cpu_load = shfind(r'[\d(?/?.\d)]+', 'cat /proc/loadavg')
    cpu_model = shfind(r'Model\s*:\s*(.*)\s*', 'cat /proc/cpuinfo')[0]

    return {
        'cpu_model': cpu_model,
        'cpu_cur_freq': float(cpu_cur_freq),
        'cpu_temp': float(cpu_temp),
        'cpu_load_1': float(cpu_load[0]),
        'cpu_load_5': float(cpu_load[1]),
        'cpu_load_15': float(cpu_load[2]),
        'running_proc': int(cpu_load[3].split('/')[1]),
    }


@register_stats
def memory(meta=None):
    mem = psutil.virtual_memory()
    return {
        'mem_available': float(str(mem.available).replace('L', '')),
        'mem_used': float(str(mem.used).replace('L', '')),
        'mem_total': float(str(mem.total).replace('L', '')),
        'mem_percent': float(mem.percent)
    }

@register_stats
def git(meta=None, root=None):
    local_hash = reip.util.shell.git('rev-parse HEAD')
    remte_hash = reip.util.shell.git('rev-parse origin/master')
    return {
        'branch': reip.util.shell.git('rev-parse --abbrev-ref HEAD'),
        'commit_date': reip.util.shell.git('log -1 --format=%ci'),
        'commit': reip.util.shell.git('rev-parse --short HEAD'),
        'uptodate': int(local_hash == remte_hash)
    }

# network

@register_stats
def wifi(wlan='wlan*', meta=None):
    iwc = ixconfig.Iwc().ifaces(wlan)
    if not iwc:
        return
    wlan = iwc[max(iwc)]
    return ({
        'wifi_quality': float(wlan['quality_ratio']) if 'quality_ratio' in wlan else None,
        'wifi_strength': float(wlan['strength']) if 'strength' in wlan else None,
        'ap': netswitch.Wpa().ssid
    })


@register_stats
def cellular(cell_tty_commands='/dev/ttyUSB2', meta=None):
    try:
        return {"cell_strength": netswitch.cell.signal_strength(cell_tty_commands)}
    except OSError:
        return


_IP = {'ip': 'inet'}
_IPMAC = {'ip': 'inet', 'mac': 'ether'}
DEFAULT_IFCONFIG = {'wlan0': _IPMAC, 'eth0': _IPMAC, 'tun0': _IP}
# DEFAULT_IFACES = ['wlan0', 'eth0', 'tun0']
# IFACE_KEY_DEFAULTS = {'ip': 'inet', 'mac': 'ether'}

# def from_defaults(defaults, a=None, default_keys=None, default_value=None, **kw):
#     if default_keys and not a and not kw:
#         a = default_keys
#     return dict(
#         ((k, defaults.get(k, default_value or k)) for k in a), **kw)

@register_stats
def network(*a, meta=None, **kw):
    cfg = dict(((k, DEFAULT_IFCONFIG.get(k, _IP)) for k in a), **kw) if a or kw else DEFAULT_IFCONFIG
    ifaces = ifcfg.interfaces()
    # wlan, tun, eth = (ifaces.get(i, {}) for i in ('wlan0', 'tun0', 'eth0'))
    return {
        'RX_packets': int(str(psutil.net_io_counters().bytes_recv).replace('L', '')),
        'TX_packets': int(str(psutil.net_io_counters().bytes_sent).replace('L', '')),
        **{
            '{}_{}'.format(pat, kname): ifaces.get(pat, {}).get(key)
            for pat, keys in cfg.items()
            # for name, ifcfg in ifaces.items()
            # if fnmatch.fnmatch(name, pat)
            for kname, key, in keys.items()
        }
    }


TYPES = {'bool': bool, 'int': int, 'str': str, 'float': float, '': lambda x: x}
def _search_usb(devices, pattern, cast=None):
    # find match
    match = next((
        d['name'] for d in devices
        if re.search(pattern, d['name'])), None)
    # cast to a type?
    for t in (cast or '').split('|'):
        match = TYPES[t](match)
    return match

@register_stats
def usb(meta=None, **devices):
    if not devices:
        return {}
    found_devices = reip.util.shell.lsusb()
    return {
        k: _search_usb(found_devices, **as_kw(kw, 'pattern'))
        for k, kw in devices.items()
    }


DEFAULT_STORAGE_LOCATIONS = ['/', '/tmp', '/var/log']

@register_stats
def storage(*poslocs, meta=None, literal_keys=False, **locs):
    # e.g.: {'/': 0.95, '/var/log': 0.45} if literal_keys else {root_usage: 0.95, varlog_usage: 0.45}
    poslocs = DEFAULT_STORAGE_LOCATIONS if not poslocs and not locs else poslocs
    locs.update({p if literal_keys else p.replace('/', '') or 'root': p for p in poslocs})
    if not literal_keys:
        locs = {'{}_usage'.format(k): p for k, p in locs.items()}

    return {k: psutil.disk_usage(path).percent for k, path in locs.items()}


def base(meta=None):
    return {
        'time': datetime.utcnow().isoformat(),
        'fqdn': socket.getfqdn(),# 'hostname'
    }

def meta(meta=None):
    return meta or {}
meta_ = meta

def full(include_meta=False):
    return reip.util.mergedict(base, cpu, memory, network, wifi, usb, storage, meta if include_meta else {})



'''Constants that can be used as sentinel values inside blocks.

'''

# this is arbitrary, just something that someone wouldn't realistically want
# to use as a value.
# normally, I'd use `object()`, but that won't work with `x is TOKEN` if it
# gets pickled.
make_token = lambda x: f'|(((( reip {x.upper()} ))))|'

# __SPLIT = '@#$%^&*()'
# _TOKEN_START, _TOKEN_END = make_token(__SPLIT).split(__SPLIT)
# is_token = lambda x: isinstance(x, str) and x.startswith(_TOKEN_START) and x.endswith(_TOKEN_END)

class Token:
    def __init__(self, name):
        self.name = name.upper()
        self.id = make_token(name)

    def __repr__(self):
        return self.id

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.check(other)

    def check(self, other):
        return isinstance(other, Token) and self.id == other.id


UNSET = Token('UNSET')
# Use this to signify a blank output. If you have a block that outputs
# [0, 1, BLANK, 2, 3], the next block would receive process(1, 2, 3, 4)
BLANK = Token('BLANK')
# return this from a block and the sources won't be incremented.
RETRY = Token('RETRY')
NEXT = Token('NEXT')
# if a block returns these, it will close/terminate itself and will
# cascade down to other blocks.
CLOSE = Token('CLOSE')
TERMINATE = Token('TERMINATE')

SOURCE_TOKENS = {RETRY, NEXT}

MIN_SLEEP = 1e-6



import collections
import numpy as np


class OnlineStats:
    '''Online statistics - calculate and update mean and variance as new data comes in.'''
    count = mean = _var = 0
    def __init__(self, history=0):
        self.last = 0
        self.samples = collections.deque(maxlen=history) if history else None

    def __getstate__(self):  # don't copy samples when pickling
        return dict(self.__dict__, samples=None)

    def __str__(self):
        return '(total = {total:.4f} avg = {mean:.6f} ± {std:.6f}, n = {count:8,})\n'.format(
            total=self.sum, mean=self.mean, std=self.std, count=self.count)

    def __len__(self):
        return self.count

    def append(self, x):
        '''Append a value to the rolling mean.'''
        # online mean
        # online std deviation
        last_mean, self.mean = self.mean, (self.mean * self.count + x) / (self.count + 1)
        if self.count:
            self._var = self._var + (x - last_mean) * (x - self.mean)

        self.last = x
        self.count += 1
        if self.samples is not None:
            self.samples.append(x)

    def extend(self, xs):
        '''Append multiple values.'''
        for x in xs:
            self.append(x)

    @property
    def sum(self):
        '''Get the rolling sum.'''
        return self.mean * self.count

    @property
    def std(self):
        '''Get the rolling standard deviation.'''
        return np.sqrt(self.var)

    @property
    def var(self):
        '''Get the rolling variance.'''
        return self._var / (self.count - 1) if self.count > 1 else 0



import time
import queue
import warnings
import numpy as np
import reip


class Rebuffer(reip.Block):
    '''Collect numpy arrays and convert them to a different resolution'''
    def __init__(self, size=None, duration=None, *a, sr_key='sr', **kw):
        assert size or duration, 'You must specify a size or duration.'
        self.size = size
        self.duration = duration
        self.sr_key = sr_key
        super().__init__(**kw)

    def init(self):
        self._q = queue.Queue()
        self.sr = None
        self.current_size = 0

    def _calc_size(self, meta):
        # calculate the size the first time using the sr in metadata
        if not self.size and meta:
            self.sr = self.sr or meta[self.sr_key]
            self.size = self.duration * self.sr

    def process(self, x, meta):
        self._calc_size(meta)

        # place buffer into queue. If buffer is full, gather buffers and emit.
        if self._place_buffer(x, meta):
            self.log.debug(f"full: {self.current_size} >= {self.size}")
            return self._gather_buffers()

    def _place_buffer(self, x, meta):
        '''put items on the queue and return if the buffer is full.'''
        xsize = self._get_size(x)
        self.current_size += xsize
        self._q.put((x, meta, xsize))
        return self.current_size >= self.size

    def _gather_buffers(self):
        '''Read items from the queue and emit once it's reached full size.'''
        size = 0
        xs, metas = [], []
        while size < self.size:
            x, meta, xsize = self._q.get()
            xs.append(x)
            metas.append(meta)
            size += xsize
            self.current_size -= xsize
        return [self._merge_buffers(xs)], self._merge_meta(metas)

    def _get_size(self, buf):
        '''Return the size of a buffer, based on it's type.'''
        return len(buf)

    def _merge_buffers(self, bufs):
        return np.concatenate(bufs)

    def _merge_meta(self, metas):
        return metas[0]


class FastRebuffer(reip.Block):
    def __init__(self, size=None, **kw):
        assert size , 'You must specify a size.'
        self.size = size
        super().__init__(**kw)

    def init(self):
        self.buffers, self.meta = None, None
        self.current_pos = 0

    def process(self, x, meta):
        # calculate the size the first time using the sr in metadata
        if self.buffers is None:
            shape = list(x.shape)
            shape[0] = shape[0] * self.size
            # self.buffers = np.concatenate([np.zeros_like(x)] * self.size)
            self.buffers = np.zeros(tuple(shape), dtype=x.dtype)
            self.meta = dict(meta)
            self.current_pos = 0

        l, i = x.shape[0], self.current_pos
        
        self.buffers[i*l:(i+1)*l, ...] = x[...]
        self.current_pos = (self.current_pos + 1) % self.size

        if self.current_pos == 0:
            ret = [self.buffers], self.meta
            self.buffers, self.meta = None, None
        else:
            ret = None

        return ret


class GatedRebuffer(reip.Block):
    '''Rebuffer while also sampling sparsely in time. For example, sampling audio files 
    while leaving gaps between them for privacy reasons.'''
    TIME_KEY = 'time'
    def __init__(self, sampler, size=10, **kw):
        self.sampler = (
            sampler if callable(sampler) else
            (lambda: sampler) if sampler else None)
        self.size = size
        self.pause_until = 0
        self.allow_index = None
        super().__init__(**kw)

    def check_input_skip(self, meta):
        if not self.sampler:
            return

        t0 = meta.get(self.TIME_KEY) or time.time()

        # check if we are currently allowing buffers
        if self.allow_index is not None:
            self.log.debug(f'allow: {self.processed} - {self.allow_index} >= {self.size}')
            if self.processed - self.allow_index < self.size:  # under the limit
                return
            # we've exceeded
            self.allow_index = None
            s = self.sampler()
            self.pause_until = t0 + s
            self.log.debug(f'will be sleeping for {s:.3g}s')

        # check if we should be pausing
        if t0 < self.pause_until:
            return True

        # if not, record the processed count
        self.allow_index = self.processed
        self.log.debug(f'allow_index: {self.allow_index}')

    def init(self):
        self.buffers = []
        self.meta = []

    def finish(self):
        self.clear()

    def clear(self):
        self.buffers.clear()
        self.meta.clear()

    def process(self, x, meta):
        if self.check_input_skip(meta):
            return
        self.log.debug(f'collecting buffer: {x.shape}')
        self.buffers.append(x)
        self.meta.append(meta)

        if len(self.buffers) >= self.size:
            X = np.concatenate(self.buffers)
            meta = self.meta[0]
            self.clear()
            self.log.debug(f'submitting: {X.shape}')
            return [X], meta




def temporal_coverage(clip_duration=10, coverage=0.5, min_silence=5.0, sampling='normal'):
    '''Given a clip_duration and desired coverage fraction, compute
    how long the silence between clips needs to be to satisfy the desired
    coverage and add some random variance to the silence based so that it
    gives values sampled from the distribution specified by sampling.

    Arguments:
        clip_duration (float): Duration of non-silent clip in seconds
        coverage (float): Fraction of time to be recorded (non-silence), must
            be in range (0,1].
        min_silence (float): Minimum silence allowed between clips, in seconds
        sampling (str): The distribution from which silence durations will be
            sampled, must be 'uniform' or 'normal'.

    Returns:
        silence (float): The amount of silence to insert between the current
            and next clip, in seconds.

    '''
    if clip_duration <= 0:
        raise ValueError('Clip duration must be positive.')
    if not (0 < coverage <= 1):
        raise ValueError('Coverage outside the allowed range of (0,1].')
    if sampling not in ['normal', 'uniform']:
        raise ValueError('Unknown sampling method.')

    # Compute exact silence duration
    silence = (1 - coverage) / float(coverage) * clip_duration

    # If the silence required to obtain the specified coverage is shorter
    # than min_silence we default back to min_silence and report the
    # estimated coverage
    if silence < min_silence:
        warnings.warn(
            "Coverage too high to meet min_silence of {:.2f} seconds, "
            "returning {:.2f}. Estimated coverage is {:.2%}".format(
                min_silence, min_silence,
                clip_duration / float(clip_duration + min_silence)))
        return min_silence

    # Add some variance
    sigma = np.abs(silence - min_silence)
    max_silence = silence + sigma

    if sampling == 'uniform':
        silence += sigma * (np.random.random_sample() * 2 - 1) # +[-sig, +sig]
    elif sampling == 'normal':
        silence += sigma / 3. * np.random.randn() # +N(0, sig/3)

    # Make sure silence is within limits
    return min(max(silence, min_silence), max_silence)



import os
import io
import base64
import tarfile
from Crypto import Random
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Util import Padding

import reip


class TwoStageEncrypt(reip.Block):
    '''Encrypt a file with a fresh generated key, then encrypt the key with
    the root key. Then tar the encrypted file and the encrypted key together.
    '''
    #PADDING = b'{'
    BLOCK_SIZE = 32

    def __init__(self, filename, rsa_key, remove_files=False, extra_files=None, is_rsa_key_file=True, **kw):
        super().__init__(**kw)
        self.filename = str(filename)
        self.remove_files = remove_files
        self.gz = self.filename.endswith('.gz')
        self.extra_files = extra_files

        if is_rsa_key_file:
            if not os.path.isfile(rsa_key):
                raise OSError("No rsa key found.")
            with open(rsa_key, 'r') as f:
                rsa_key = f.read()
        self.public_key = RSA.importKey(str(rsa_key))
        self.cipher = PKCS1_OAEP.new(self.public_key)


    def process(self, file, meta):
        '''Encrypt file with AES 4096.
        Arguments:
            file: name of the file to be encrypted
            path: Location to save encrypted file
            default: current dir
        '''
        fname = reip.util.ensure_dir(self.filename.format(
            name=reip.util.fname(file), fname=os.path.basename(file), **meta))

        with open(file, 'rb') as f:
            enc_message = f.read()

        enc = self.encrypt(enc_message, os.path.basename(file))

        extra_files = reip.util.resolve_call(self.extra_files, file, meta=meta) or {}
        enc.update(extra_files)

        compressed = self.compress(enc, fname)
        self.maybe_remove_files(file)
        return [compressed], {}

    def encrypt(self, msg, name='file'):
        # Encrypt the file using AES and AES using RSA
        aes_key = Random.new().read(AES.key_size[2])
        key = AES.new(aes_key, AES.MODE_CBC)

        msg = Padding.pad(msg, AES.block_size)
        msg = key.iv + key.encrypt(msg)
        msg = base64.b64encode(msg)

        # gather files
        return {
            name+'.enc': msg,
            name+'.key': self.cipher.encrypt(aes_key)
        }

    def compress(self, files, out_file):
        # write tar
        with tarfile.open(out_file, 'w:gz' if self.gz else 'w') as tar:
            for f, data in files.items():
                tar_addbytes(tar, f, data)
        return out_file

    def maybe_remove_files(self, *files):
        if self.remove_files:
            for f in files:
                os.remove(f)


class TwoStageDecrypt(reip.Block):
    '''Decrypt a file encrypted with TwoStepEncryptFile.
    '''
    PADDING = b'{'
    BLOCK_SIZE = 32

    def __init__(self, filename, rsa_key, remove_files=False, **kw):
        super().__init__(**kw)
        self.filename = str(filename)
        self.remove_files = remove_files

        if os.path.isfile(rsa_key):
            with open(rsa_key, 'r') as f:
                rsa_key = f.read()
        self.private_key = RSA.importKey(rsa_key)
        self.cipher = PKCS1_OAEP.new(self.private_key)

    def process(self, file, meta=None):
        # get the output filename and make sure the parent dir exists
        fname = reip.util.ensure_dir(self.filename.format(
            name=reip.util.fname(file), **meta))
        # decrypt file to disk
        enc_msg, enc_key = self.decompress(file)
        decrypted = self.decrypt(enc_msg, enc_key)
        with open(fname, 'wb') as f:
            f.write(decrypted)

        self.maybe_remove_files(file)
        return [fname], {}

    def decrypt(self, msg, enc_key):
        '''Decrypt file with AES 4096.'''
        msg = base64.b64decode(msg)

        iv, msg = msg[:AES.block_size], msg[AES.block_size:]
        aes_key = self.cipher.decrypt(enc_key)
        key = AES.new(aes_key, AES.MODE_CBC, iv)

        msg = key.decrypt(msg)
        msg = Padding.unpad(msg, AES.block_size)
        return msg

    def decompress(self, filename):
        # write file and key to tar file
        with tarfile.open(filename, 'r') as tar:
            return (
                next(tar.extractfile(f).read() for f in tar.getmembers()
                     if f.name.endswith('.enc')),
                next(tar.extractfile(f).read() for f in tar.getmembers()
                     if f.name.endswith('.key'))
            )

    def maybe_remove_files(self, *files):
        if self.remove_files:
            for f in files:
                os.remove(f)


def get_private_key(public_fname):
    private_fname = public2private(public_fname)
    return private_fname if os.path.isfile(private_fname) else public_fname


def public2private(public_fname):
    return '{}_private{}'.format(*os.path.splitext(public_fname))

def private2public(private_fname):
    return '{}.pub'.format(private_fname)

def load_rsa(rsa_key):
    with open(rsa_key, 'rb') as f:
        return RSA.importKey(f.read())

def create_rsa(private_fname=None, public_fname=None, bits=2048):
    private_key = RSA.generate(bits, e=65537)
    public_key = private_key.publickey()

    if not (private_fname or public_fname):
        private_fname = os.path.expanduser('~/.ssh/encrypt_rsa')
    private_fname = private_fname or public2private(public_fname)
    public_fname = public_fname or private2public(private_fname)

    reip.util.ensure_dir(public_fname)
    with open(public_fname, 'wb') as f:
        f.write(public_key.exportKey("PEM"))
    with open(private_fname, 'wb') as f:
        f.write(private_key.exportKey("PEM"))
    return private_fname, public_fname

def tar_addbytes(tar, f, data):
    t = tarfile.TarInfo(f)
    t.size = len(data)
    tar.addfile(t, io.BytesIO(data))


from .pointers import Pointer
from .customer import Customer


class BaseStore:
    '''A base class for Stores.'''
    Pointer = Pointer
    Customer = Customer

    def __str__(self):
        return '<{} n={}>'.format(self.__class__.__name__, len(self))

    def __len__(self):
        raise NotImplementedError

    def put(self, data, meta=None, id=None):
        raise NotImplementedError

    def get(self, id):
        raise NotImplementedError

    def delete(self, ids):
        raise NotImplementedError



'''




'''
import time
from contextlib import contextmanager
import weakref
import multiprocessing as mp
import remoteobj

import reip
from reip.util import text, iters


def default_graph():
    return _ContextScope.default

def get_graph(graph=None):
    return _ContextScope.default if graph is None else graph

def top_graph():
    return _ContextScope.top


class _ContextScope:
    top = None
    default = None
    all_instances = {}



class BaseContext:  # (metaclass=_MetaContext)
    _delay = 1e-3
    _previous = False  # the previous default instance
    parent_id, task_id = None, None

    def __init__(self, *blocks, name=None, graph=None):
        if blocks and not name and isinstance(blocks[0], str):
            name, blocks = blocks[0], blocks[1:]
        self.blocks = list(blocks)
        self.name = name or f'{self.__class__.__name__}_{id(self)}'
        self.parent_id, self.task_id = BaseContext.register_instance(self, graph)

        for b in self.blocks:
            self.register_instance(b, self)

    # global instance management

    @classmethod
    def _initialize_default_graph(cls):
        _ContextScope.default = _ContextScope.top = Graph(name='_default_')

    @classmethod
    def get(cls, instance=None):
        '''If `instance` is `None`, the default instance will be returned.
        Otherwise `instance` is returned.'''
        return _ContextScope.default if instance is None else instance

    def add(self, block):
        '''Add block to graph.'''
        self.blocks.append(block)

    def remove(self, block):
        self.blocks.remove(block)

    def clear(self):
        self.blocks.clear()

    @classmethod
    def register_instance(cls, child=None, parent=None):
        '''Add a member (Block, Task) to the graph in instance.

        This is used inside of

        Arguments:
            child (reip.Graph, reip.Block): A graph, task, or block to add to
                instance.
                If `member is instance`, nothing will be added. In other words,
                A graph cannot be added to itself.
            parent (reip.Graph): A graph or task to be added to.
                If instance is None, the default graph/task will be used.
                If instance is False, nothing will be added.

        Returns:
            parent_id (str): the name of the child's parent block.
                Can be a Graph or Task.
            task_id (str or None): the name of the task that a child is attached to.
                If the parent graph is not inside a task or is not a task itself,
                it will return None.

            NOTE: Returning string ids prevents Blocks from having a circular reference
            to the entire graph.

        Basically this should handle:
         - adding a graph to a graph
            - parent.name, parent.task_id
         - adding a task to a graph
            - raise if parent is task
            - raise if graph.task_id is not None
            - if `flatten_tasks`, then iterate through graph parents until you
              find one on the main task
            - parent.name, child.name
         - adding a graph to a task
            - parent.name, parent.task_id
         - adding a block to a graph
            - parent.name, parent.task_id
         - adding a block to a task
            - parent.name, parent.task_id
         - do nothing if parent is False / None+unset default.
        '''
        _ContextScope.all_instances[child.name] = weakref.ref(child)
        task_id = child.name if isinstance(child, reip.Task) else None

        # user explicitly disabled graph for this child
        if parent is False:
            return None, task_id
        parent = get_graph(parent)  # get default if None
        # there is no graph specified and no default graph
        if parent is None:
            return None, task_id

        # trying to add to self
        if parent is child:
            return parent.parent_id, task_id or parent.task_id

        if task_id:  # disallow nested tasks
            # parent_contexts = [parent] + parent.parent_contexts if flatten_tasks else [parent]
            if isinstance(parent, reip.Task):
                raise RuntimeError(
                    'Cannot add a task ({}) to another task ({}).'.format(
                        child.name, parent.name))
            if parent.task_id:
                raise RuntimeError(
                    'Cannot add a task ({}) to a graph ({}) in a different task ({}).'.format(
                        child.name, parent.name, parent.task_id))

        # everything checks out.
        parent.add(child)
        return parent.name, task_id or parent.task_id

    @classmethod
    def get_object(cls, id, require=True):
        '''Get an instance using its name. If the instance '''
        obj = _ContextScope.all_instances.get(id) if id is not None else None
        obj = obj() if obj is not None else None
        if obj is None and require:
            raise ValueError(f'Object {id} does not exist.')
        return obj

    @property
    def parents(self):
        obj = self
        stack = []
        while obj.parent_id:
            obj = BaseContext.get_object(obj.parent_id)
            if obj is None:
                break
            stack.append(obj)
        return stack

    def __enter__(self):
        return self.as_default()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.restore_previous()

    def as_default(self):
        '''Set this graph as the default graph and track the previous default.'''
        self._previous, _ContextScope.default = _ContextScope.default, self
        return self

    def restore_previous(self):
        '''Restore the previously set graph.'''
        if self._previous is not False:
            _ContextScope.default, self._previous = self._previous, None
        self._previous = False
        return self

    # def rename_child(self, old_name, new_name):
    #     pass


# import blessed

class Graph(BaseContext):
    '''Represents a collection of Blocks.
    
    
    '''
    _delay = 1e-3
    _main_delay = 1e-2#0.1
    controlling = None

    def __init__(self, *blocks, log_level='info', **kw):
        super().__init__(*blocks, **kw)
        self.log = reip.util.logging.getLogger(self, level=log_level)
        self._except = remoteobj.LocalExcept()

    def __repr__(self):
        return '{}({}), {} children:{}\n'.format(
            self.__class__.__name__, self.name, len(self.blocks),
            ''.join('\n' + text.indent(b) for b in self.blocks))

    def __bool__(self):
        return not (self.done or self.error)

    @classmethod
    def detached(cls, *blocks, **kw):
        return cls(*blocks, graph=None, **kw)

    # run graph

    def run(self, duration=None, stats_interval=1, print_graph=True, **kw):
        '''Run a graph and all of its child blocks.
        
        .. code-block::

            with reip.Graph() as g:
                BlockA().to(BlockB()).to(BlockC())

            g.run(duration=10)
        '''
        if print_graph:
            print("\nStarting:", self, "\n")

        with self.run_scope(**kw):
            self.wait(duration, stats_interval=stats_interval)

    @contextmanager
    def run_scope(self, wait=True, raise_exc=True):
        '''Run a graph and its child blocks while inside the context manager.
        
        .. code-block:: python

            with reip.Graph() as g:
                BlockA().to(BlockB()).to(BlockC())

            with g.run_scope():
                # gives you more flexibility in here
                g.wait(duration=5)
            # equivalent to g.run(duration=5)
        '''
        self.log.debug(text.green('Starting'))
        # controlling = False
        try:
            with self._except('spawn', raises=False):
                try:
                    self.spawn(wait=wait)
                    # controlling = self.controlling
                    self.log.debug(text.green('Ready'))
                except Exception as e:
                    self.log.error(text.red('Spawn Error'))
                    self.log.exception(e)
                    raise
            with self._except(raises=False):
                yield self
        except KeyboardInterrupt as e:
            self.log.warning(text.yellow('Interrupting'))
            # reip.util.print_stack('Interrupted here')
            self.log.exception(e)
            self.terminate()
        finally:
            try:
                self.join(raise_exc=raise_exc)
            finally:
                self.log.debug(text.green('Done'))
                # if controlling:
                #     print(self.stats_summary())

    # _delay = 1
    def wait(self, duration=None, stats_interval=None):
        t, t0 = time.time(), time.time()
        for _ in iters.timed(iters.sleep_loop(self._delay), duration):
            if self.done:
                return True

            if stats_interval and time.time() - t > stats_interval:
                t = time.time()
                status = self.status()
                self.log.info("Status after %.3f sec:\n%s" % (time.time() - t0, status[status.find("\n")+1:]))

    def _reset_state(self):
        self._except.clear()

    # state

    @property
    def ready(self):
        '''True if all child blocks are ready'''
        return all(b.ready for b in self.blocks)

    @property
    def running(self):
        '''True if any child blocks are running'''
        return any(b.running for b in self.blocks)

    @property
    def terminated(self):
        '''True if any child blocks are ready'''
        return all(b.terminated for b in self.blocks)

    @property
    def done(self):
        '''True if any child blocks are done'''
        return any(b.done for b in self.blocks)

    @property
    def error(self):
        '''True if any child blocks threw an error'''
        return bool(self._except.all()) or any(b.error for b in self.blocks)

    # Block control
    _ready_flag = None
    def spawn(self, wait=True, _controlling=True, _ready_flag=None, **kw):
        '''Spawns the graph and its children.'''
        self.controlling = _controlling
        if self.controlling:
            self._ready_flag = _ready_flag = mp.Event()

        self._spawn_tasks_first(_ready_flag=_ready_flag, **kw)
        for block in self.blocks:
            block.spawn(wait=False, _controlling=False, _ready_flag=_ready_flag, **kw)

        if wait:
            self.wait_until_ready()
        if self.controlling:
            self.raise_exception()
            if self._ready_flag is not None:
                _ready_flag.set()

    def _spawn_tasks_first(self, **kw):
        for b in self.blocks:
            if isinstance(b, reip.Task):
                b.spawn(wait=False, _controlling=False, **kw)
            elif isinstance(b, reip.Graph):
                b._spawn_tasks_first(wait=False, _controlling=False, **kw)

    def wait_until_ready(self, stats_interval=5):
        t = time.time()
        while not self.ready and not self.done:
            time.sleep(self._delay)
            
            if stats_interval and time.time() - t > stats_interval:
                t = time.time()
                self.log.info(f'waiting for all blocks to initialize...\n{self}')

    def join(self, close=True, terminate=False, raise_exc=None, **kw):
        '''Joins the graph and its children.'''
        if self._ready_flag is not None and not self._ready_flag.is_set():
            self._ready_flag.set()
        if close:
            self.close()
        if terminate:
            self.terminate()
        for block in self.blocks:
            block.join(terminate=False, raise_exc=False, **kw)
        if raise_exc is None and self.controlling:
            raise_exc = True
        if raise_exc:
            self.raise_exception()
        self.controlling = self._ready_flag = None

    def pause(self):
        '''Pause all blocks.'''
        for block in self.blocks:
            block.pause()

    def resume(self):
        '''Resume all blocks.'''
        for block in self.blocks:
            block.resume()

    def close(self):
        '''Close all blocks.'''
        for block in self.blocks:
            block.close()

    def terminate(self):
        '''Terminate all blocks.'''
        for block in self.blocks:
            block.terminate()

    def raise_exception(self):
        '''Raise any exceptions from blocks.'''
        for block in self.blocks:
            block.raise_exception()
        self._except.raise_any()

    def __export_state__(self):
        return {
            'blocks': [b.__export_state__() for b in self.blocks],
            '_except': self._except,
        }

    def __import_state__(self, state):
        if state:
            for b, update in zip(self.blocks, state.pop('blocks', ())):
                b.__import_state__(update)

    def short_str(self):
        return '[{}({})[{} children]]'.format(
            self.__class__.__name__[0],
            self.name, len(self.blocks))

    def stats(self):
        return {
            'name': self.name,
            'blocks': {b.name: b.stats() for b in self.blocks}
        }

    def summary(self):
        return '\n'.join(s for s in (b.summary() for b in self.blocks) if s)

    def status(self):
        return text.b_(
            f'[{self.name}]',
            text.indent('\n'.join(
                s for s in (b.status() for b in self.blocks) if s)), ""
        )

    def stats_summary(self):
        return text.block_text(
            f'[{self.name}]',
            *(s for s in (b.stats_summary() for b in self.blocks) if s)
        )


# create an initial default graph
# Graph.top = Graph.default = Graph()



import time
import fnmatch
import warnings
import logging
import numpy as np
import gi
gi.require_version('Gst', '1.0')
gi.require_version("GstApp", "1.0")
gi.require_version('GstVideo', '1.0')
from gi.repository import Gst, GstApp, GLib, GObject, GstVideo


class GStream:
    _inited = False

    @classmethod
    def initialize(cls):
        GObject.threads_init()
        Gst.init(None)
        cls._inited = True

    _bus = None
    done = False
    _debug = True
    _state_timeout = 1e+5
    def __init__(self, auto_play=True, pipeline=None, logger=None):
        if not self._inited:
            warnings.warn('GStreamer is uninitialized. Initializing automatically.')
            self.initialize()

        self._pipeline = pipeline or Gst.Pipeline()
        self.auto_play = auto_play
        self.log = logger or logging.getLogger(__name__)

    def __len__(self):
        return len(self._pipeline.children)

    def __iter__(self):
        return (p.name for p in self._pipeline.children[::-1])

    def __getitem__(self, key):
        return (
            self._pipeline.get_child_by_index(key) if isinstance(key, int) else
            self._pipeline.get_child_by_name(key))

    @property
    def children(self):
        return self._pipeline.children

    # Graph Definition

    def add(self, name, title=None, **kw):
        '''Add an element.'''
        element = (
            name if isinstance(name, Gst.Element) else
            Gst.ElementFactory.make(name, title))
        if element is None:
            raise RuntimeError("Could not create element: " + name)

        self._pipeline.add(element)
        for k, v in kw.items():
            element.set_property(k.replace('_', '-'), v)
            # Gst.caps_from_string(v) if isinstance(v, str) else v
        return element

    def link(self, *elements, start=None, end=None):
        '''Connect consecutive elements together.'''
        # select starting element
        if not elements:
            elements = tuple(self)
            elements = elements[
                _get_element_index(elements, start):
                _get_element_index(elements, end)]

        elements = [
            el if isinstance(el, (Gst.Element, Gst.Pad)) else self[el]
            for el in elements]

        # add elements to gst object
        # for e in elements:
        #     self._pipeline.add(e)

        # connect elements together
        for el1, el2 in zip(elements, elements[1:]):
            if not el1.link(el2):
                raise RuntimeError("Could not link elements: {} -> {}".format(
                    el1.name, el2.name))
        return self

    def addcap(self, capstr, title=None, **kw):
        return self.add(
            'capsfilter', title,
            caps=cap(capstr) if isinstance(capstr, str) else capstr,
            **kw)

    def search(self, *patterns):
        '''Get all element names matching patterns.'''
        return [
            k for k in self
            if any(fnmatch.fnmatch(k, p) for p in patterns)
        ]

    def find(self, *patterns):
        '''Get all elements whose name matches patterns.'''
        return [self[k] for k in self.search(*patterns)]

    # Control Flow

    @property
    def ready(self):
        # Gst.State.READY is little different from Gst.State.NULL, so
        # we judge whether pipeline is ready by Gst.State.PAUSED or Gst.State.PAYLING
        return self._check_state(Gst.State.PLAYING, Gst.State.PAUSED)

    @property
    def running(self):
        return self._check_state(Gst.State.PLAYING)

    def _check_state(self, *states):
        return self._pipeline.get_state(self._state_timeout).state in states

    def toggle(self, running=None, msg1='resume', msg2='pause'):
        running = not self.running if running is None else running
        state, msg = (Gst.State.PLAYING, msg1) if running else (Gst.State.PAUSED, msg2)
        if self.running != running:
            if self._pipeline.set_state(state) == Gst.StateChangeReturn.FAILURE:
                if msg:
                    raise RuntimeError("Could not {} pipeline".format(msg))
        return True

    def start(self, running=True):
        self.done = False
        self._bus = self._pipeline.get_bus()
        self.toggle(running, 'start')

    def pause(self):
        return self.toggle(False)

    def resume(self):
        return self.toggle(True)

    def end(self):
        self._pipeline.send_event(Gst.Event.new_eos())

    def check_messages(self, all_pending=True):
        if not self._bus:
            raise ValueError("Invalid bus")

        while self._bus.have_pending():
            message = self._bus.pop()
            if not message:
                raise ValueError("Empty message")

            t = message.type
            if t == Gst.MessageType.ERROR:
                err, debug = message.parse_error()
                raise RuntimeError((
                    "Error received from element {}: {}\n"
                    "Debugging information:{}").format(
                        message.src.get_name(), err, debug))

            elif t == Gst.MessageType.EOS:
                if self._debug:
                    self.log.debug("End-Of-Stream reached")
                self._pipeline.set_state(Gst.State.NULL)
                time.sleep(1e-3)  # Bus stops receiving messages after EOS for some reason
                self.done = True

            elif t == Gst.MessageType.STATE_CHANGED:
                if isinstance(message.src, Gst.Pipeline):
                    old, new, pending = message.parse_state_changed()
                    self.log.debug(
                        "Pipeline state changed from {} to {}".format(
                            old.value_nick, new.value_nick))
            else:
                self.log.debug('Unknown message: {}'.format(message.type))
            if not all_pending:
                break



def _get_element_index(elements, search):
    if search is None or isinstance(search, int):
        return search
    try:
        return next(
            len(elements) - i
            for i, el in enumerate(tuple(elements)[::-1], 1)
            if el == search or elements[el] is search)
    except StopIteration:
        raise ValueError('No element matching {}'.format(search))


def cap(x):
    return Gst.caps_from_string(x)

def element(x, *a, **kw):
    return Gst.ElementFactory.make(x, *a, **kw)

def pad(name, src=None):
    return Gst.Pad.new(
        name,
        Gst.PadDirection.SRC if src else
        Gst.PadDirection.UNKNOWN if src is None else
        Gst.PadDirection.SINK)


def unpack_sample(sample, fmt=None):
    buf = sample.get_buffer()
    caps = sample.get_caps().get_structure(0)
    sz, imfmt = buf.get_size(), caps.get_value('format')
    w, h = caps.get_value('width'), caps.get_value('height')
    assert not fmt or imfmt == fmt

    X = np.ndarray(sz, buffer=buf.extract_dup(0, sz), dtype=np.uint8)
    split = int(len(X) // (w*h) * (w*h))
    X, extra = X[:split], X[split:]
    X = X.reshape((h, w, -1))
    ts = buf.pts
    return X, ts



import os


def setonce(file, func):
    file = os.path.expanduser(file)
    if os.path.isfile(file):
        with open(file, 'r') as f:
            return f.read().rstrip('\n')
    value = str(func())
    os.makedirs(os.path.dirname(os.path.abspath(file)), exist_ok=True)
    with open(file, 'w') as f:
        f.write(value)
    return value



import time
from contextlib import contextmanager
import reip
from reip.util import text
import remoteobj


class BaseAgent:
    _delay = 1e-6
    ready = done = terminated = False
    _EXCEPT_CLASS = remoteobj.LocalExcept

    def __init__(self, name=None, graph=None):
        self.name = name or f'{self.__class__.__name__}_{id(self)}'
        self.parent_id, self.task_id = reip.Graph.register_instance(self, graph)
        self.log = reip.util.logging.getLogger(self)
        self._except = self._EXCEPT_CLASS()

    def _reset_state(self):
        self._except.clear()




    def run(self, duration=None, **kw):
        with self.run_scope(**kw):
            self.wait(duration)

    @contextmanager
    def run_scope(self, wait=True, raise_exc=True):
        try:
            with self._except('spawn', raises=False, log=self.log):
                self.spawn(wait=wait)
            yield self
        except KeyboardInterrupt:
            self.log.info(text.yellow('Interrupting'))
            self.terminate()
        finally:
            self.join(raise_exc=raise_exc)

    def wait(self, duration=None):
        for _ in reip.util.iters.timed(reip.util.iters.sleep_loop(self._delay), duration):
            if self.done or self.error:
                return True

    def wait_until_ready(self):
        while not self.ready and not self.error and not self.done:
            time.sleep(self._delay)



    def spawn(self, wait=True):
        self.ready = True

    def join(self, raise_exc=True):
        self.done = True

    def terminate(self):
        self.terminated = True



    @property
    def _exception(self):
        return self._except.last

    @property
    def error(self):
        return self._exception is not None

    def raise_exception(self):
        self._except.raise_any()

    def log_exception(self):
        for e in self._except.all():
            self.log.exception(e)




    def __export_state__(self):
        return self.__dict__

    def __import_state__(self, state):
        self.__dict__.update(state)



'''File System Events

Watchdog API Reference: https://pythonhosted.org/watchdog/api.html

.. note::

    from experience, I've found file watchers can be a bit unreliable - (potentially
    due to my own misuse !).

    Additionally, many times (for example file upload) watching file 
    events may not actually be what you want, and you may be better off
    polling file lists using glob as that will reduce the number of missed files
    and is resilient to reboots.

'''
import queue
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
import reip
from reip.util import text

class _WatchBlockHandler(PatternMatchingEventHandler):
    def __init__(self, q, event_types, *a, **kw):
        self.q = q
        self.event_types = event_types
        super().__init__(*a, **kw)

    def on_any_event(self, event):
        try:
            if not self.event_types or event.event_type in self.event_types:
                self.q.put(event)
        except Exception as e:
            self.q.put(e)

class Watch(reip.Block):
    '''A Generic File Watcher'''
    _q = _event_handler = _watch = None
    event_types = None
    def __init__(self, *patterns, path='./', event_types=None, recursive=False, **kw):
        self.patterns = list(patterns or ('*',))
        self.path = path
        self.event_types = event_types or self.event_types
        self.recursive = recursive
        super().__init__(n_inputs=0, **kw)

    _Handler = _WatchBlockHandler

    # define a global observer - is this right?

    _observer = None
    @property
    def observer(self):
        if Watch._observer is None:
            Watch._observer = Observer()
            Watch._observer.start()
        return Watch._observer

    @observer.setter
    def observer(self, value):
        Watch._observer = value

    def init(self):
        # a queue is used to gather all events emitted
        self._q = queue.Queue()
        # create watchdog stuff
        self._event_handler = self._Handler(
            self._q, self.event_types, patterns=self.patterns)
        self._watch = self.observer.schedule(
            self._event_handler, self.path, recursive=self.recursive)

    def process(self, meta):
        if not self._q.empty():
            e = self._q.get()
            if isinstance(e, Exception):
                raise e
            return self._output_event(e, meta)

    def _output_event(self, event, meta):
        return [event.src_path], {'event_type': event.event_type}

    def finish(self):
        # remove handler
        self.observer.remove_handler_for_watch(self._event_handler, self._watch)
        if self.observer.is_alive() and not any(self.observer._handlers.values()):
            # if there are no more handlers, shutdown the observer
            self.observer.unschedule_all()
            self.observer.stop()
            self.observer.join()
            self.observer = None


class Created(Watch):
    '''Outputs any created files.'''
    event_types = ('created',)

class Modified(Watch):
    '''Outputs any modified files.'''
    event_types = ('modified',)

class Deleted(Watch):
    '''Outputs any deleted files.'''
    event_types = ('deleted',)

class Moved(Watch):
    '''Outputs any moved files.'''
    event_types = ('moved',)
    def _output_event(self, event, meta):
        return [event.src_path, event.dest_path], {'event_type': event.event_type}


# Watch.Created = Created
# Watch.Modified = Modified
# Watch.Deleted = Deleted
# Watch.Moved = Moved



import glob
import time
import fnmatch
import itertools
import collections
import numpy as np

import reip
from reip.util import text


class Iterator(reip.Block):
    '''Yield each element in an iterator one at a time.'''
    def __init__(self, iterator, **kw):
        self.iterator = iter(iterator)
        super().__init__(n_inputs=0, **kw)

    def process(self, meta=None):
        try:
            x = next(self.iterator)
            return [x], {}
        except StopIteration:
            return reip.CLOSE


class Interval(reip.Block):
    '''Emit every X seconds.'''
    def __init__(self, seconds=2, initial=None, **kw):
        self.seconds = seconds
        self.initial = initial
        super().__init__(n_inputs=0, max_rate=1/seconds, **kw)

    def init(self):
        time.sleep(self.initial or 0)

    def process(self, meta=None):
        return [None], {'time': time.time()}


# class ControlledDelay(reip.Block):
#     '''Add system time to metadata.'''
#     def process(self, x, meta=None):
#         time.sleep(x)
#         return [None], {'time': time.time()}


class Time(reip.Block):
    '''Add system time to metadata.'''
    def __init__(self, **kw):
        super().__init__(**kw)

    def process(self, x, meta):
        return [x], {'time': time.time()}


class Meta(reip.Block):
    '''Add arbitrary data to metadata.'''
    def __init__(self, meta, **kw):
        self.meta = meta or {}
        super().__init__(*a, **kw)
    def process(self, x, meta=None):
        return [x], {
            k: v(meta) if callable(v) else v
            for k, v in self.meta.items()
        }



class Glob(reip.Block):
    '''Outputs files matching glob patterns.'''
    def __init__(self, *paths, recursive=True, atonce=False, **kw):
        self.paths = paths
        self.recursive = recursive
        self.atonce = atonce  # return all files at once
        super().__init__(n_inputs=0, **kw)

    def init(self):
        self.last = set()

    def process(self, meta=None):
        fs = {f for p in self.paths for f in glob.glob(p, recursive=self.recursive)}
        self.last, fs = fs, fs - self.last
        if self.atonce:
            yield [fs], {}
        else:
            for f in fs:
                yield [f], {}

# as an experiment:
# @reip.helpers.asbasiccontext
# def Glob(self, *paths, atonce=False):
#     self.last = set()
#     def process():
#         fs = {f for p in paths for f in glob.glob(p)}
#         self.last, fs = fs, fs - self.last
#         if atonce:
#             yield fs
#         else:
#             yield from fs
#     yield process


# class Dict(reip.Block):
#     '''Merge block outputs into a dict.'''
#     def __init__(self, meta, *a, **kw):
#         super().__init__(*a, **kw)
#
#     def process(self, *xs, meta=None):
#         return [xs], {}

class AsDict(reip.Block):
    '''Merge block outputs into a dict.'''
    def __init__(self, *columns, prepare=lambda c, x: (c, x), meta=None, **kw):
        self.columns = columns
        self.prepare = prepare
        self.meta_keys = reip.util.as_list(meta or ())
        super().__init__(n_inputs=len(columns), **kw)

    def process(self, *xs, meta=None):
        data = {}
        # set data buffer keys
        for c, x in zip(self.columns, xs):
            c, x = self.prepare(c, x)
            if isinstance(c, (list, tuple, np.ndarray)):
                data.update(zip(c, reip.util.as_iterlike(x)))
            else:
                data[c] = x
        # set meta keys
        if self.meta_keys:
            data.update({
                k: meta[k] for k in meta
                if any(fnmatch.fnmatch(k, p) for p in self.meta_keys)
            })
        return [data], {}


# class Sleep(reip.Block):
#     def __init__(self, sleep=2, **kw):
#         self.sleep = sleep
#         super().__init__(**kw)

#     def process(self, x, meta):
#         time.sleep(self.sleep)
#         return [x], meta


class Constant(reip.Block):
    '''Yields a constant value.'''
    def __init__(self, value, *a, **kw):
        self.value = value
        super().__init__(*a, n_inputs=0, **kw)

    def process(self, meta):
        return [self.value], meta


class Increment(Iterator):
    '''Yield an incrementing (or decrementing) value.'''
    def __init__(self, start=None, stop=None, step=1, **kw):
        if stop is None:
            start, stop = 0, start
        super().__init__(
            range(start or 0, stop, step) if stop is not None else
            itertools.count(start or 0, step or 1), **kw)


class Debug(reip.Block):
    '''Debug the output of a block.'''
    level = 'debug'
    log_level = 'debug'
    def __init__(self, message=None, level=None, convert=None, value=False, compact=False,
                 summary=False, period=None, name=None, border=False, log_level=None, **kw):
        self.message = message or 'Debug'
        self.value = value
        self.period = period
        self._summary = summary
        self._border = border
        self._last_time = 0
        self.level = reip.util.logging.aslevel(level or self.level)
        self.convert = convert
        self.compact = compact
        name = 'Debug-{}'.format(message.replace(" ", "-")) if message else None
        super().__init__(name=name, log_level=log_level or self.log_level, **kw)

    def _block_line_format(self, x):
        if self.convert:
            return type(x).__name__, self.convert(x)
        if isinstance(x, np.ndarray):
            if self._summary:
                return x.shape, x.dtype, 'min:', x.min(), 'max:', x.max(), 'mean:', x.mean()
            if x.size > 40 or x.ndim > 2:
                return x.shape, x.dtype, x if self.value else '' # , f'{np.isnan(x).sum()} nans'
            return x.shape, x.dtype, x
        return type(x).__name__, x

    def _block_format(self, *xs, meta):
        data = [self._block_line_format(x) for x in xs]

        return (text.block_text if self._border else text.b_)(
            text.blue(self.message),
            *(
                (('data:', data, '||', 'meta:', meta),)
                if self.compact else
                ('data:', *[('  ',i)+tuple(d) for i,d in enumerate(data)], ('meta:', meta))
            )
        )

    def process(self, *xs, meta=None):
        if not self.period or time.time() - self._last_time > self.period:
            self._last_time = time.time()
            self.log.log(self.level, self._block_format(*xs, meta=meta))
            # print(, flush=True)
        return xs, meta


class Results(reip.Block):
    '''Gather the results in lists as self.results and self.meta'''
    squeeze = True
    def __init__(self, squeeze=True, **kw):
        self.squeeze = squeeze
        super().__init__(**kw)

    def init(self):
        self.results = []
        self.meta = []

    def process(self, *xs, meta=None):
        self.results.append(xs[0] if self.squeeze and len(xs) == 1 else xs)
        self.meta.append(meta)


class Lambda(reip.Block):
    '''An inline way to create a block from a regular function. Input and output formats are the same.'''
    def __init__(self, func=None, name=None, init=None, finish=None, n_inputs=None, **kw):
        self.func = func or (lambda *a, **kw: None)
        self.init_func = init
        self.finish_func = finish
        if not func:  # don't run too fast if there's no process function
            kw.setdefault('max_rate', 1)
        super().__init__(
            name=name or self.func.__name__, 
            n_inputs=n_inputs, **kw)

    def init(self):
        if self.init_func:
            self.init_func()

    def process(self, *xs, meta=None):
        return self.func(*xs, meta=meta)

    def finish(self):
        if self.finish_func:
            self.finish_func()

    @classmethod
    def define(cls, func):
        return cls(func)


class Interleave(reip.Block):
    '''Interleave the block inputs to a single output.'''
    def __init__(self, sort_key=None, **kw):
        super().__init__(n_inputs=None, **kw)
        self.sort_key = sort_key

    def process(self, *xs, meta=None):
        if self.sort_key:
            xs = [x for i, x in sorted(enumerate(xs), key=lambda x: meta[x[0]][self.sort_key])]
        for x in xs:
            if x is not None:
                yield [x], meta


class Separate(reip.Block):
    '''Separate out a single input into multiple outputs based on a condition per output.
    '''
    def __init__(self, bins, **kw):
        self.bins = bins
        super().__init__(n_outputs=len(bins), **kw)

    def process(self, x, meta=None):
        out = [None]*len(self.bins)
        for i, check in enumerate(self.bins):
            if check(x, meta):
                out[i] = x
                break
        return out, meta


class Gather(reip.Block):
    def __init__(self, size=None, reduce=None, squeeze=True, **kw):
        self.size = size
        self.reduce = reduce
        self.squeeze = squeeze
        super().__init__(**kw)

    def init(self):
        self.items = []

    def _should_fire(self):
        return len(self.items) >= self.size

    def process(self, *xs, meta=None):
        if self.squeeze and len(xs) == 1:
            xs = xs[0]
        self.items.append(xs)
        if self._should_fire():
            items, self.items = self.items, []
            if self.reduce:
                items = self.reduce(items)
            return [items], meta

    def finish(self):
        self.items = None



import setuptools

USERNAME = 'reip-project'
NAME = 'reip'

setuptools.setup(
    name=NAME,
    version='0.0.3',
    description='',
    long_description=open('README.md').read().strip(),
    long_description_content_type='text/markdown',
    author='Yurii Piadyk, Bea Steers',
    author_email='ypiadyk@nyu.edu',
    url='https://github.com/reip-project/reip-pipelines',
    packages=setuptools.find_packages(),
    # entry_points={'console_scripts': ['{name}={name}:main'.format(name=NAME)]},
    install_requires=[
        # cli
        'fire', 'colorlog',
        # core packages
        'numpy',
        'remoteobj>=0.3.1',
        # 'pyarrow',
        # block packages
        'watchdog',
        # 'tflit>=0.1.2',
        # status messages
        'psutil', 'ifcfg', 'ixconfig>=0.1.1', 'netswitch', 'requests',
    ],
    extras_require={
        # 'status': ['psutil', 'ixconfig', 'netswitch'],
        'plasma': ['pyarrow'],
        'tflite': ['tflit>=0.1.2'],
        'audio': ['sounddevice', 'librosa'],
        'video': ['opencv-python'],  # already installed on Jetson by default
        # 'gstream': ['PyGObject'],
        'encrypt': ['pycryptodome'],
        'vis': ['matplotlib'],
        'docs': ['sphinx!=1.3.1', 'sphinx_theme', 'sphinx_rtd_theme'],
    },
    license='BSD 3-Clause Clear License',
    keywords='iot embedded app pipeline block multiprocessing')



import numpy as np
import cv2
import reip



class OpticalFlow(reip.Block):
    '''
    https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_video/py_lucas_kanade/py_lucas_kanade.html#dense-optical-flow-in-opencv
    '''
    hsv = None
    _prev = None

    # pyr_scale=0.5, levels=3, winsize=15, iterations=3,
    # poly_n=5, poly_sigma=1.2, flags=0
    def __init__(self, pyr_scale=0.5, levels=3, winsize=15, iterations=3,
                 poly_n=5, poly_sigma=1.2, flags=0, resize=(256, 256), draw_mag_scale=10, draw=None, **kw):
        self.pyr_scale = pyr_scale
        self.levels = levels
        self.winsize = winsize
        self.iterations = iterations
        self.poly_n = poly_n
        self.poly_sigma = poly_sigma
        self.flags = flags
        self.resize = resize
        self._should_draw = draw
        self.draw_mag_scale = draw_mag_scale
        super().__init__(n_outputs=2, **kw)

    def init(self):
        self.should_draw = bool(self.sinks[1].readers) if self._should_draw is None else self._should_draw
        self.log.debug(f'will draw: {self.should_draw}')

    def process(self, frame, meta):
        original = frame
        if self.resize:
            frame = cv2.resize(frame, self.resize)
        current = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype('uint8')
        if self.hsv is None:
            self.hsv = np.zeros_like(frame)
            self.hsv[..., 1] = 255
            self._prev = current
            return

        flow = cv2.calcOpticalFlowFarneback(
            self._prev, current, None, self.pyr_scale, self.levels, self.winsize,
            self.iterations, self.poly_n, self.poly_sigma, self.flags)
        self._prev = current

        flow = cv2.cartToPolar(flow[..., 0], flow[..., 1])
        return [flow, self.draw(original, flow) if self.should_draw else None], {}

    def draw(self, im, flow, mix=0.5):
        mag, ang = flow
        self.hsv[..., 0] = ang * 180 / np.pi / 2
        # self.hsv[..., 2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
        self.hsv[..., 2] = np.clip(mag*self.draw_mag_scale, 0, 255)
        flow = cv2.cvtColor(self.hsv, cv2.COLOR_HSV2BGR)

        if im.shape != flow.shape:
            flow = cv2.resize(flow, im.shape[:2][::-1])
        im = im * (1-mix) + flow * mix
        return (im/255).astype('float32')


import time
import functools
from . import iters


def throttle(func, interval=None):
    '''Don't call a function more frequently than `duration`. Caches the
    return value from the previous call.'''
    if not interval:
        return func

    @functools.wraps(func)
    def inner(*a, **kw):
        t = time.time()
        if t - inner.t0 >= interval:
            inner.t0 = t
            inner.retval = func(*a, **kw)
        return inner.retval

    inner.t0, inner.retval = 0, None
    return inner


def retry(func, n=None, exc=Exception, log=None, **kw):
    '''retry a function on failure.'''
    @functools.wraps(func)
    def inner(*a, **kw):
        for i in iters.run_loop(n=n, **kw):
            try:
                return func(*a, **kw)
            except exc as e:
                if n is not None and i == n:
                    raise
                if log is not None:
                    log.error('({}, try={}) {}'.format(e.__class__.__name__, i+1, e))
    return inner



from .base import *
from .pointers import *
from .store import *
from .queue import *
from .customer import *
try:
    import pyarrow
except ImportError:
    PlasmaStore = ArrowQueue = None
    HAS_PYARROW = False
else:
    from .plasma import *
    HAS_PYARROW = True
from .queue_store import *
from .producer import *



import csv
import reip
import numpy as np


class _CsvWriter(reip.util.CycledWriter):
    _file = None
    def __init__(self, filename, headers=None, max_rows=60, **kw):
        self.headers = headers
        self._kw = kw
        super().__init__(filename, file_length=max_rows)

    def _new_writer(self, fname, meta):
        self._file = open(fname, 'w', newline='')

        writer = csv.writer(self._file, **self._kw)
        if self.headers:
            writer.writerow(self.headers)
        return writer

    def _close_writer(self):
        if self._file is not None:
            self._file.close()

class Csv(reip.Block):
    '''A CSV writer.'''
    _fname = _file = _writer = None
    def __init__(self, filename='{time}.csv', headers=None, max_rows=1000, **kw):
        self.writer = _CsvWriter(filename, headers=headers, max_rows=max_rows)
        super().__init__(**kw)

    def process(self, X, meta):
        for x in np.atleast_2d(X):
            writer = self.writer.get(meta)
            writer.writerow(list(x))
            self.writer.increment(1)
        self.writer.get(meta)
        return self.writer.output_closed_file()

    def finish(self):
        self.writer.close()



import os
import numpy as np
import librosa
import reip
import reip.blocks as B
import reip.blocks.ml



as_model_file = lambda f: os.path.join(os.path.dirname(reip.__file__), 'models', f)

class EdgeL3(B.ml.Tflite):
    model_path = as_model_file('edgel3_embedding.tflite')
    model_params = dict(
        target_sr=8000, 
        duration=1, 
        hop_size=0.1, 
        n_fft=1024,
        hop_length=160,
        n_mels=64, 
    )

    def __init__(self, model_path=None, **kw):
        super().__init__(model_path or self.model_path, **kw)

    def process(self, data, meta):
        if isinstance(data, str):
            y, sr = librosa.load(data, sr=None)
        else:
            y, sr = data, meta.get('sr')
        S = preprocess_audio(y, sr, **dict(self.model_params, **self.extra_kw))
        z = self.model(S)
        return [z], meta


class EdgeL3Embedding2Class(B.ml.Tflite):
    classes = [
        'engine',
        'machinery_impact',
        'non_machinery_impact',
        'powered_saw',
        'alert_signal',
        'music',
        'human_voice',
        'dog',
    ]
    model_path = as_model_file('edgel3_mlp_classification.tflite')
    def __init__(self, model_path=None, **kw):
        super().__init__(model_path or self.model_path, **kw)






def preprocess_audio(y, sr, duration, hop_size, target_sr=None, **kw):
    '''Get the audio data at a specified target sample rate.'''
    if y.ndim == 2:
        y = y[:,0]
    # proper sr
    y = librosa.resample(y, orig_sr=sr, target_sr=target_sr or sr)
    sr = target_sr or sr
    # frame
    framelen = int(sr * duration)
    hoplen = int(sr * hop_size)
    y = librosa.util.pad_center(y, size=framepadlen(len(y), framelen, hoplen))
    frames = librosa.util.frame(y, frame_length=framelen, hop_length=hoplen)
    # build mel spec
    S = np.array([
        melstft(frame, sr, **kw)
        for frame in frames.T
    ])[...,None]
    return S


def framepadlen(xlen, flen, hlen):
    '''Calculate the padding for a specific frame and hop size.'''
    return int(np.ceil(1. * (xlen - flen) / hlen) * hlen + flen)


def melstft(X, sr, n_fft=2048, hop_length=512, n_mels=64):
    # magnitude spectrum
    S = np.abs(librosa.core.stft(
        X, n_fft=n_fft, hop_length=hop_length,
        window='hann', center=True, pad_mode='constant'))

    # log mel spectrogram
    return librosa.power_to_db(
        librosa.feature.melspectrogram(
            S=S, sr=sr, n_mels=n_mels, htk=True),
        amin=1e-10)



# def ml_stft_inputs(data, meta, hop_size=0.1, duration=1, sr=8000, **kw):
#     # load audio
#     file, y = (data, None) if isinstance(data, str) else (None, data)
#     y, sr = load_resample(file, y, meta.get('sr'), sr)
#     # frame and extract stft
#     frames = padframe(y, int(sr * duration), int(sr * hop_size)).T
#     X = npgenarray(
#         (melstft(frame, sr, **kw) for frame in frames),
#         len(frames))[..., None]
#     return X

# def melstft(X, sr, n_fft=2048, hop_length=512, n_mels=64):
#     # magnitude spectrum
#     S = np.abs(librosa.core.stft(
#         X, n_fft=n_fft, hop_length=hop_length,
#         window='hann', center=True, pad_mode='constant'))

#     # log mel spectrogram
#     return librosa.power_to_db(
#         librosa.feature.melspectrogram(
#             S=S, sr=sr, n_mels=n_mels, htk=True),
#         amin=1e-10)




# '''

# Utils

# '''

# def load_resample(fname=None, y=None, sr=None, target_sr=None):
#     '''Get the audio data at a specified target sample rate.'''
#     if y is None:
#         y, target_sr = librosa.load(fname, mono=False, sr=target_sr)
#     else:
#         y = y.T
#         y = np.atleast_2d(y)[0, :]
#         if len(y) == 1:
#             y = y[0]
#         if sr and target_sr and sr != target_sr:
#             y = librosa.resample(y, sr, target_sr)
#     return y, target_sr or sr


# def framepadlen(xlen, flen, hlen):
#     '''Calculate the padding for a specific frame and hop size.'''
#     return int(np.ceil(1. * (xlen - flen) / hlen) * hlen + flen)


# def padframe(y, framelen, hoplen):
#     '''Get framed array with zero padding to fill the first/last frames.'''
#     y = librosa.util.pad_center(y, framepadlen(len(y), framelen, hoplen))
#     return librosa.util.frame(y, framelen, hoplen)



import reip
import numpy as np


# class BeatTracker(reip.util.statistics.OnlineStats):
#     last_time = None
#     def append(self, t):
#         self.last_time, last = t, self.last_time
#         if last is not None:
#             return super().append(t - last)


class Synchronize(reip.Block):
    def __init__(self, tol=1e-1, key='time', **kw):
        '''Synchronize two Block outputs'''
        self.key = key
        self.tol = tol
        super().__init__(n_inputs=None, n_outputs=None, **kw)

    # XXX: so that the number of sinks always equals the number of sources
    def set_block_source_count(self, n):
        super().set_block_source_count(n)
        self.set_block_sink_count(n)

    def process(self, *xs, meta):
        # if any queue slot is empty, retry
        is_none = [x is None for x in xs]
        if any(is_none):
            return reip.RETRY

        # check the time differences between samples
        # if any are too far back in the past, then skip them (???)
        # FIXME: is this right??
        times = np.array([m[self.key] for m in meta.sources])
        i_lead = np.argmax(times)
        isbehind = (times[i_lead] - times) > self.tol
        if np.any(isbehind):
            return [reip.NEXT if b else reip.RETRY for b in isbehind], meta
        return xs, meta



import reip
import multiprocessing as mp
from reip.stores import Store, PlasmaStore, QueueStore, Pointer, Counter, HAS_PYARROW


class Producer(reip.Sink):
    def __init__(self, size=100, delete_rate=10, task_id=reip.UNSET, skip_no_readers=True, **kw):
        self.size = size + 1  # need extra slot because head == tail means empty
        self.delete_rate = delete_rate
        self.task_id = task_id
        self.stores = {}
        self.head = Pointer(self.size)
        self.tail = Pointer(self.size)
        self.skip_no_readers = skip_no_readers
        # self._dropped = Counter()
        self.readers = []
        super().__init__(**kw)

    def __str__(self):
        return '<{} \n  head={}\n  tail={}\n  readers={}\n  stores={}>'.format(
            self.__class__.__name__, self.head, self.tail,
            [str(p) for p in self.readers],
            [str(s) for s in self.stores.values()])

    def spawn(self):
        for store in self.stores.values():
            if hasattr(store, 'spawn'):
                store.spawn()

    def join(self):
        for store in self.stores.values():
            if hasattr(store, 'join'):
                store.join()

    def __len__(self):
        return self.head.counter - min(
            (r.counter for r in self.readers),
            default=self.tail.counter)

    def full(self):
        self._trim()
        return (self.head.counter - self.tail.counter) >= (self.size - 1)

    # @property
    # def dropped(self):
    #     return self._dropped.value
    #
    # @dropped.setter
    # def dropped(self, value):
    #     self._dropped.value = value

    def _trim(self):
        '''Delete any stale items from the queue.'''
        if self.readers:
            # get the number of stale entries
            new_value = min(
                (r.counter for r in self.readers),
                default=self.tail.counter)
            to_delete = [
                v % self.size for v in range(self.tail.counter, new_value)]

            # delete items if there's enough
            if len(to_delete) > self.size / self.delete_rate:
                # print("Trimming", len(to_delete))
                for store in self.stores.values():
                    store.delete(to_delete)
                self.tail.counter = new_value

    def _put(self, buffer):
        '''Send data to stores.'''
        if self.skip_no_readers and not self.readers:
            return
        data, meta = buffer
        meta = meta or {}
        for store in self.stores.values():
            store.put(data, meta, id=self.head.pos)
        self.head.counter += 1

    def gen_source(self, task_id=reip.UNSET, throughput='small', **kw):
        '''Generate a source from this sink.

        Arguments:
            task_id (str): the identifier for the task.
            throughput (str): The size of data. This determines the serialization
                method to use. Should be:
                 - 'small' for data < 1GB
                 - 'medium' for data < 3GB
                 - 'large' for data > 3GB
            **kwargs: arguments to pass to `store.Customer`.
        '''
        # create the store if it doesn't exist already
        same_context = (
            reip.UNSET.check(task_id) or reip.UNSET.check(self.task_id) or
            task_id == self.task_id)
        store_id = same_context, throughput
        if store_id not in self.stores:  # use True/False as store keys
            if same_context:
                store = Store(self.size)
            else:
                if HAS_PYARROW and throughput == 'large':
                    store = PlasmaStore(self.size)
                else:
                    store = QueueStore(self.size)
                    if HAS_PYARROW and throughput == 'medium':
                        kw.setdefault('serializer', 'arrow')

                # convert pointers to shared pointers
                self.head = self.head.as_shared()
                self.tail = self.tail.as_shared()
                # self._dropped = self._dropped.as_shared()
            self.stores[store_id] = store

        # create a customer and a pointer.
        store = self.stores[store_id]
        self.readers.append(store.Pointer(self.size, self.tail.counter))
        return store.Customer(self, len(self.readers) - 1, store_id=store_id, **kw)



'''
Manage a list of functions/context managers as if they are one.
'''
import time
import reip


def lineprofile(block, run):
    try:
        import pyinstrument
        prof = pyinstrument.Profiler()
        prof.start()
        return run()
    finally:
        prof.stop()
        print(prof.output_text(unicode=True, color=True))


def reboot(block, run):
    while not block.closed:
        try:
            return run()
        except Exception as e:
            block.log.error(reip.util.excline(e))


def retry(n=None, sleep=None):
    def retry(block, run):
        for i in reip.util.iters.loop():
            try:
                return run()
            except Exception as e:
                if n and i == n-1:
                    raise
                block.log.error(reip.util.excline(e))
                if sleep:
                    time.sleep(sleep)
            else:
                break
    return retry


def suppress(block, run):
    try:
        return run()
    except Exception:
        pass


def heartrate(block, run):
    import heartrate
    heartrate.trace(browser=True)
    return run()




# Text formatting

class C:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def _text_wrapper(start, end=C.ENDC):
    return '{}{{}}{}'.format(start, end).format

red = _text_wrapper(C.FAIL)
blue = _text_wrapper(C.OKBLUE)
green = _text_wrapper(C.OKGREEN)
yellow = _text_wrapper(C.WARNING)
bold = _text_wrapper(C.BOLD)
underline = _text_wrapper(C.UNDERLINE)

t4 = ' '*4
t2 = ' '*2

def printasone(*txt, end='\n', **kw):
    print(' '.join(map(str, txt)) + end, end='', **kw)

# block text helpers

def indent(x, n=1, w=4, ch=' '):
    '''Indent text using spaces.'''
    return ''.join(ch * w * n + l for l in str(x).splitlines(keepends=True))

def tabindent(x, n=1, w=4):
    '''Indent text using tabs.'''
    return indent(space2tab(x, w=w), w=1, n=n, ch='\t')


def tab2space(text, w=4):
    '''Convert tabs to spaces.'''
    return text.replace('\t', ' '*w)

def space2tab(text, w=4):
    '''Convert spaces to tabs.'''
    return text.replace(' '*w, '\t')


def trim_indent(text, tw=2):
    '''Remove any common indent from text.

    Arguments:
        text (str): the text to re-indent.
        tw (int): the number of spaces per tab character.
    '''
    # normalize tabs to spaces
    lines = text.replace('\t', ' '*tw).splitlines()
    # get the min indent to norm to
    m = min([len(l) - len(l.lstrip()) for l in lines if l.strip()], default=0)
    # rejoin the text with the proper indent
    return '\n'.join([l[m:] for l in lines])


def striplines(text):
    '''Like text.strip() but it only removes lines with purely whitespace and
    leaves text indentation.'''
    lines = text.splitlines()
    i, j = 0, len(lines)
    while not lines[i].strip():
        i += 1
    while not lines[j-1].strip():
        j -= 1
    return '\n'.join(lines[i:j])

def strip_each_line(txt):
    return '\n'.join(l.rstrip() for l in txt.splitlines())

def comment(txt, ch='#', n=1, spaces=1):
    '''Apply prefix to each line. Defaults to python comments.'''
    ch, spaces = ch*n, " "*spaces
    return '\n'.join('{}{}{}'.format(ch, spaces, l) for l in txt.splitlines())


def block_text(*txts, n=20, ch='*', div=''):
    '''Create a block of text with a character border.'''
    return ch * n + '\n{}\n'.format(comment(b_(*txts, div=div), ch=ch)) + ch * n

def b_(*lines, div=''):
    '''Convert arguments to lines. Lines can be tuples (will be joined by a space.)'''
    div = '\n{}\n'.format(div) if div else '\n'
    return div.join(
        l_(*l) if isinstance(l, (list, tuple)) else str(l)
        for l in lines if l is not None)

def l_(*line, div=' '):
    '''Convert arguments to a space separated string'''
    return div.join(map(str, line))


def fw_(*line, w=20, right=False):
    '''As a fixed width string'''
    return f"{l_(*line):{'>' if right else '<'}{w}}"

def tbl(*rows, buffer=2):
    '''Format as a table. Calculates column widths.'''
    rows = [[str(c) for c in cs] for cs in rows]
    widths = [max((len(c) for c in cs), default=0) for cs in zip(*rows)]
    return b_(*([fw_(c, w=w + buffer-1) for c, w in zip(cs, widths)] for cs in rows))


if __name__ == '__main__':
    print(red('this should be red'))
    print(blue('this should be blue'))
    print(green('this should be green'))
    print(yellow('this should be yellow'))
    print(underline('this should be underline'))

    print()
    print(b_(
        ('hi', blue('im blue'), 56),
        'okay',
        '',
        red('bloop'),
    ))

    print()
    print(block_text(
        l_('hiiii', red('okay')),
        bold('well'),
        green('alright'),
    ))



import re
import subprocess
import shlex
import collections
import reip

ShellResult = collections.namedtuple('ShellResult', 'out err rc cmd')


def run(cmd, *a, **kw):
    '''Run a shell command. Dictionary arguments passed will be converted to
    bash flags. e.g.: dict(x=5, y=True, z=None) -> '-x 5 -y'

    Arguments:
        cmd (str): the command to run. By default, arguments will be quoted.
            To pass a value without quoting, use the format pattern `{!r}`.
        *args, **kwargs: arguments to format command with.
            If any arg is None, it will insert an empty string.
            If any argument is a dict, it will attempt to format it as bash flags.
             - If the key is a single character, only a single preceding dash
               will be used.
             - if the value is True, then it will output like a boolean flag.
               e.g.: dict(asdf=True) => --asdf
             - if the value is None or False, then it will be omitted.
             - otherwise, it will be cast to a string.

    Examples:
    >>> shell.run('echo 10')  # echo 10
    # ('10\n', '')
    >>> shell.run('echo {}', '10 && echo 15')  # echo '10 && echo 15'
    # ('10 && echo 15\n', '')
    >>> shell.run('echo {!r}', '10 && echo 15')  # echo 10 && echo 15
    # ('10\n15\n', '')
    >>> shell.run('echo {} {b} {a}', 10, a=11, b=15)  # echo 10 15 11
    # ('10 15 11\n', '')
    >>> iface = None  # wlan0
    ... shell.run('ping {} 8.8.8.8', dict(I=iface, c=3))  # ping -c 3 8.8.8.8
    # ('PING 8.8.8.8 ...', '')
    # NOTE: notice how because -I is None, it gets filtered out.

    '''
    cmd = build(cmd, *a, **kw)
    r = subprocess.run(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        check=False, shell=True)

    return ShellResult(r.stdout.decode('utf-8'), r.stderr.decode('utf-8'), r.returncode, cmd)


def build(cmd, *a, **kw):
    return cmd.format(
        *(ShellArg(x) for x in a),
        **{k: ShellArg(v) for k, v in kw.items()})


class ShellArg:
    '''Formats a user-specified argument as a bash argument.

    If value is a dict, it will be constructed as a series of flags. Otherwise
    it will just convert to string and potentially quote the value.
    '''
    def __init__(self, value):
        self.value = value

    def _format(self, quote=False):
        if self.value is None:
            return ''
        if isinstance(self.value, dict):
            return ' '.join(
                '{}{} {}'.format(
                    '--' if len(k) > 1 else '-', k,
                    '' if v is True else shlex.quote(str(v)) if quote else v
                ).strip()
                for k, v in self.value.items()
                if v is not None and v is not False
            )
        return shlex.quote(str(self.value)) if quote else str(self.value)

    def __repr__(self):
        return self._format(quote=False)

    def __str__(self):
        return self._format(quote=True)


# Misc Commands

def shmatch(cmd, out=None, err=None, rc=None):
    '''
    >>> shmatch('piwatcher status', 'OK')
    >>> shmatch('ifconfig wlan0', 'inet [.\d]+')
    >>> shmatch('docker logs blah --tail 100', err='Error')
    >>> shmatch('ls')  # true
    >>> shmatch('ls', 'README')  # true
    >>> shmatch('ls 1>&2', 'README')  # false
    >>> shmatch('ls 1>&2', err='README')  # true
    >>> shmatch('true')  # true
    >>> shmatch('true', rc=1)  # false
    >>> shmatch('false')  # false
    >>> shmatch('false', rc=1)  # true

    '''
    result = run(cmd)
    if out is None and err is None and rc is None:
        rc = 0
    return (
        (rc is None or result.rc == rc) and
        (out is None or doesmatch(result[0], out)) and
        (err is None or doesmatch(result[1], err)) or False)

def doesmatch(txt, pat=None):
    if pat is None:
        return False
    if isinstance(pat, list):
        return any(re.search(p, txt) for p in pat)
    if isinstance(pat, dict):
        return next(k for k, p in pat.items() if re.search(p, txt))
    return bool(re.search(pat, txt))  # TODO: maybe return match(es)?


def git(*cmd, root=None):
    '''Run a git command in the sonycnode repository.'''
    return run('git', dict(C=root), *cmd).out.strip()

LSUSB_FMT = (
    r'Bus\s+(?P<bus>\d+)\s+Device\s+(?P<device>\d+).+ID\s(?P<id>\w+:\w+)\s(?P<name>.+)\s*$')
def lsusb():
    res = run('lsusb')
    return [] if res.err else [
        dict(d, dev='/dev/bus/usb/{}/{}'.format(d['bus'], d['device']))
        for d in (reip.util.matchmany(l, LSUSB_FMT) for l in res.out.splitlines())
    ]


def default_routes(first=True):
    res = run('route')
    routes = [x.split()[-1] for x in res.out.splitlines() if x.startswith('default')]
    return routes and routes[0] or None if first else routes
